@@ -1,75 +1,51 @@
 #!/usr/bin/env python
 '''Utilities for list and tuple manipulation by David Mashburn.
-Notably:
-    totuple -> recursive conversion of any nested iterable to tuple
-    makeTuple -> like totuple, but for non-iterables, returns (a,)
-    iterToX -> like totuple, but for any conversion
-    intOrFloat -> converts string to either int if possible, float otherwise
+Most notably:
     flatten -> drop 1 (or more if specified) levels of nesting from a data structure
     zipflat -> returns the flattened form of a zip
-    removeDuplicates -> order-perserving, recursive duplicate-remover
+    ziptranspose -> pure-function version of zip(*_)
+    deletecases -> remove all occurrences of each of cases from a list
+    removeDuplicates -> order-perserving, duplicate-remover
     partition -> partition a list into n-sized chunks
+    roll -> list version of numpy.roll
+    interp -> floating point "indexing" using linear interpolation when necessary
+    
+    totuple -> recursive conversion of any nested iterable to tuple
+    makeTuple -> like totuple, but for non-iterables, returns (a,)
+    getMaxDepth -> get the maximum depth level of a nested list
+    iterToX -> like totuple, but for any conversion
+    
     fancyIndexingList (alias fL) -> powerful numpy-like indexing for lists,
                                     use like fl(someList)[someIndexingStuff]
-    fancyIndexingListM1 (alias fLm1) -> like fL, but subtracts 1 from all indices recursively'''
+    fancyIndexingListM1 (alias fLm1) -> like fL, but subtracts 1 from all indices recursively
+    
+    There are other functions here as well, but these are the most useful/novel'''
 
 import operator
 
-def totuple(a):
-    '''Makes tuples out of nested datastructures like lists and arrays.
-       Authored by Bi Rico, http://stackoverflow.com/questions/10016352/convert-numpy-array-to-tuple'''
-    try:
-        return tuple(totuple(i) for i in a)
-    except TypeError: # dig until we can dig no more!
-        return a
-def makeTuple(a):
-    '''Like totuple, but ensures that you get a tuple out.'''
-    retVal = totuple(a)
-    return ( retVal if retVal.__class__==tuple else (retVal,) )
-def iterToX(f,iter):
-    '''Replace all iterables in a nested structure (list,tuple,etc...)
-       with another type "f" (actually just any function returning iterables)
-       Like totuple, but more general, also uses list comprehensions
-       instead of generators for more generality (notably sympy.Tuple)'''
-    try:
-        return f(*[iterToX(f,i) for i in iter])
-    except TypeError: # dig until we can dig no more!
-        return iter
-def intOrFloat(string):
-    '''Not sure if your string is formatted as an int or a float? Use intOrFloat instead!'''
-    try:
-        return int(string)
-    except ValueError:
-        return float(string)
-def floatIntStringOrNone(s):
-    '''An even more generic version of intOrFloat... Might be better to use try-except instead of this, but it works for now...'''
-    if s=='None':
-        return None
-    dotC,eC,EC = s.count('.'),s.count('e'),s.count('E')
-    weirdDashCount = ( 0 if len(s)==1 else (s[1:].replace('e-','').replace('E-','')).count('-') )
-    if dotC>1 or eC>1 or EC>1 or not s.replace('e','').replace('E','').replace('.','').replace('-','').isdigit():
-        print 'Error! invalid format!'
-        return s
-    if dotC>0 or eC>0 or EC>0:
-        return float(s)
-    else:
-        return int(s)
+###############################
+## Some basic list utilities ##
+###############################
+
 def flatten(l,repetitions=1):
-    '''A wrapper around the generator-based list flattener (quite fast)'''
+    '''A wrapper around the generator-based list flattener (quite fast)
+       
+       Use like: flatten(aList,3)
+       
+       Another (nominally faster) version of this is: [ l for i in aList for j in i for k in j for l in k ]
+       but it's unable to deal with jagged arrays...
+       
+       A terse but slower version of flatten (1 repetition) can also be acheived with: sum(aList,[])'''
     retVal = l
     for i in range(repetitions):
         gen = ( ( k if hasattr(k,'__iter__') else [k] ) for k in retVal ) # Makes the function able to deal with jagged arrays as well
         retVal = [j for i in gen for j in i]
     return retVal
-# Use like: flatten(aList,3)
-
-# Another (nominally faster) version of this is: [ l for i in aList for j in i for k in j for l in k ]
-# but it's unable to deal with jagged arrays...
-# A terse but slower version of flatten (1 repetition) can also be acheived with: sum(aList,[])
 
 def zipflat(*args):
     '''Like zip, but flattens the result'''
     return [j for i in zip(*args) for j in i]
+
 def ziptranspose(l):
     '''Tranpose the two outer dimensions of a nest list
        ( This is just a wrapper around zip(*l) )'''
@@ -80,15 +56,20 @@ def removeDuplicates(l):
        Modified version of code found here: http://stackoverflow.com/questions/480214/how-do-you-remove-duplicates-from-a-list-in-python-whilst-preserving-order'''
     seen = set()
     return [ x for x in totuple(l) if x not in seen and not seen.add(x)]
+
+def removeAdjacentDuplicates(l):
+    '''Replace any groups of items with the same value with a single occurrence instead.
+       Ex:
+       removeAdjacentDuplicates([1,2,3,0,0,1,2,0,0,1,1,1,2]) --> [1,2,3,0,1,2,0,1,2]'''
+    return [ l[i] for i in range(len(l)-1)
+                  if l[i]!=l[i+1] ] + l[-1:]
+
 def deletecases(l,cases):
     '''Delete all elements of list "cases" from list "l"'''
     if not hasattr(cases,'__iter__'):
         cases=[cases]
-    l=copy(l) # work on a copy
-    for c in cases:
-        for i in range(l.count(c)):
-            l.remove(c)
-    return l
+    return [ i for i in l if i not in cases ]
+
 def partition(l,n,clip=True):
     '''Partition list "l" into "n"-sized chunks
        clip chops off whatever does not fit into n-sized chunks at the end'''
@@ -103,13 +84,87 @@ def roll(l,n=1):
     else:
         return l
 
+def groupByFunction(l,f,appendFun=None):
+    '''Break up a list into groups (a dict of smaller lists) based on a
+       common property (the result of the function f applied to each element).
+       Optionally, transform the elements before appending them with appendFun.
+       
+       Examples:
+       >>> groupByFunction([1,2,3,4,5,6,7,8],lambda x:x<5)
+       {False: [5, 6, 7, 8], True: [1, 2, 3, 4]})
+       >>> groupByFunction([(1,1),(1,2),(-1,3),(-1,4)],lambda x:x[0])
+       {1: [(1, 1), (1, 2)], -1: [(-1, 3), (-1, 4)]}
+       >>> groupByFunction([(1,1),(1,2),(-1,3),(-1,4)],lambda x:x[0],lambda x:x[1])
+       {1: [1, 2], -1: [3, 4]}
+       
+       where obviously the lambdas in the last 2 examples could be
+       replaced with calls to operator.itemgetter instead
+       
+       This method is based on an example given for collections.defaultdict in:
+       http://docs.python.org/2/library/collections.html
+       '''
+    groupDict = {}
+    for i in l:
+        groupDict.setdefault(f(i),[]).append(appendFun(i) if appendFun else i)
+    return groupDict
+
+def interp(l,index):
+    '''Basically floating point indexing with interpolation in between.'''
+    m = index % 1
+    indexA=int(index)
+    if m==0:
+        return l[indexA]
+    else:
+        indexB = indexA + (1 if index>=0 else -1)
+        return l[indexA]*(1-m) + l[indexB]*(m)
+
+##########################################
+## Some utilities for nested structures ##
+##########################################
+
+def totuple(a):
+    '''Makes tuples out of nested datastructures like lists and arrays.
+       Authored by Bi Rico, http://stackoverflow.com/questions/10016352/convert-numpy-array-to-tuple'''
+    try:
+        return tuple(totuple(i) for i in a)
+    except TypeError: # dig until we can dig no more!
+        return a
+
+def makeTuple(a):
+    '''Like totuple, but ensures that you get a tuple out.'''
+    retVal = totuple(a)
+    return ( retVal if retVal.__class__==tuple else (retVal,) )
+
+def iterToX(f,iterable):
+    '''Replace all iterables in a nested structure (list,tuple,etc...)
+       with another type "f" (actually just any function returning iterables)
+       Like totuple, but more general, also uses list comprehensions
+       instead of generators for more generality (notably sympy.Tuple)'''
+    try:
+        return f(*[iterToX(f,i) for i in iterable])
+    except TypeError: # dig until we can dig no more!
+        return iterable
+
 def getMaxDepth(l,depth=0):
     '''Get the maximum depth of any nested structure.
         For a numpy array, this is the same as ndim.'''
     if not hasattr(l,'__iter__'):
         return depth
     else:
-        return max([getRecursiveDepth(i,depth+1) for i in l])
+        return max([getMaxDepth(i,depth+1) for i in l])
+
+def getBoundingShape(l):
+    '''Get the minimum shape of an ND-array that will hold this structure.
+       For a numpy array, this is the same as shape.
+       This function can only be called on iterables and will fail otherwise;
+       (atomic objects have no shape)'''
+    shape = []
+    childShapes = [ getBoundingShape(i) for i in l if hasattr(i,'__iter__') ]
+    if len(childShapes)>0:
+        maxLen = max([ len(s) for s in childShapes ])
+        for i in range(maxLen):
+            shape.append( max( [ s[i] for s in childShapes if len(s)>i] ) )
+    return (len(l),)+tuple(shape)
 
 def replaceNodesWithNone(l):
     '''A useful way to present the structure of nested objects.
@@ -120,9 +175,11 @@ def replaceNodesWithNone(l):
 def applyAtNodes(f,l,*args,**kdws):
     return [ ( applyAtNodes(f,i,*args,**kdws) if hasattr(i,'__iter__') else f(i,*args,**kdws) )
              for i in l ]
+
 def applyAtAllDepths(f,l,*args,**kdws):
     return f([ f( applyAtAllDepths(f,i,*args,**kdws) if hasattr(i,'__iter__') else f(i,*args,**kdws) )
              for i in l ])
+
 def applyAtDepth(f,l,depth,*args,**kdws):
     '''Apply a unary function to any nested structure at a certain depth
         With depth=0, this  is just f(l)).'''
@@ -137,7 +194,7 @@ def applyAtDepth(f,l,depth,*args,**kdws):
 
 def applyInfix_ShallowCompare(f,x,y,depth=0):
     '''Apply any function f to the elements of 2 nested list stuctures,
-        (like numpy does with + on arrays, but more general).
+        (like numpy does with "+" on arrays, but more general).
         Unlike numpy. this defaults to trying to match shallowest structure first:
             a(f,[A,B],[[1,2],[3,4]]) => [[f(A,1),f(A,2)],[f(B,3),f(B,4)]]
         For this to work, x and y must have identical structures down to the point
@@ -158,7 +215,9 @@ def applyInfix_ShallowCompare(f,x,y,depth=0):
             return [a(f,x,i,depth+1) for i in y]
         else:
             return f(x,y)
+
 applyInfix = applyInfix_ShallowCompare
+
 def applyInfix_DeepCompare(f,x,y,depth=0,xStructure=None,yStructure=None):
     '''Apply any function f to the elements of 2 nested list stuctures,
         (like numpy does with "+" on arrays, but more general).
@@ -180,7 +239,7 @@ def applyInfix_DeepCompare(f,x,y,depth=0,xStructure=None,yStructure=None):
         else:
             raise TypeError("Nested structure mismatch at depth "+str(depth)+"!")
     elif depthX==0 or depthY==0:
-        return applyFunctionToNestedStructureA(f,x,y,depth+1)
+        return applyInfix_ShallowCompare(f,x,y,depth+1)
     elif depthX<depthY:
         xStructure = (replaceNodesWithNone(x) if xStructure==None else xStructure)
         return [a(f,x,i,depth+1,xStructure=xStructure) for i in y]
@@ -190,22 +249,15 @@ def applyInfix_DeepCompare(f,x,y,depth=0,xStructure=None,yStructure=None):
 
 def shallowAdd(x,y):
     return applyInfix_ShallowCompare(operator.add,x,y)
+
 def shallowMul(x,y):
     return applyInfix_ShallowCompare(operator.mul,x,y)
+
 def deepAdd(x,y):
     return applyInfix_DeepCompare(operator.add,x,y)
-def deepAdd(x,y):
-    return applyInfix_DeepCompare(operator.mul,x,y)
 
-def interp(l,index):
-    '''Basically floating point indexing with interpolation in between.'''
-    m = index % 1
-    indexA=int(index)
-    if m==0:
-        return l[indexA]
-    else:
-        indexB = indexA + (1 if index>=0 else -1)
-        return l[indexA]*(1-m) + l[indexB]*(m)
+def deepMul(x,y):
+    return applyInfix_DeepCompare(operator.mul,x,y)
 
 def interpGen(l,index):
     '''Just like interp except that it uses the generic shallowAdd and shallowMul.'''
