@@ -1,10 +1,19 @@
 #!/usr/bin/env python
 '''Utilities for array and list manipulation by David Mashburn.
 Notably:
-    functions to scale arrays by integer multiples (without interpolation)
-    drawing basic graphics objects on (multi-dimensional) numpy arrays:
-       line segment, triangle, circle, and sphere
-'''
+    shape_multiply, shape_multiply_zero_fill ->
+        functions to scale arrays by integer multiples (without interpolation)
+    
+    polyArea -> get polygon area from points
+    
+    BresenhamFunction, BresenhamTriangle ->
+        draw lines and planes in N-dimensional space
+    
+    ImageCircle, ImageSphere ->
+        draw circles and spheres (2D or 3D)
+    
+    The drawing functions only use integer arithmetic and return a
+    list of the coordinates that can be as array indices'''
 
 import numpy as np
 from copy import copy
@@ -14,6 +23,11 @@ from list_utils import *
 
 one = np.array(1) # a surprisingly useful little array; makes lists into arrays by simply one*[[1,2],[3,6],...]
 
+def multidot(*args):
+    '''Multiply multiple arguments with np.dot:
+       reduce(np.dot,args,1)'''
+    return reduce(np.dot,args,1)
+
 def limitInteriorPoints(l,numInteriorPoints,uniqueOnly=True):
     '''return the list l with only the endpoints and a few interior points (uniqueOnly will duplicate if too few points)'''
     inds = np.linspace(0,len(l)-1,numInteriorPoints+2).round().astype(np.integer)
@@ -21,14 +35,15 @@ def limitInteriorPoints(l,numInteriorPoints,uniqueOnly=True):
         inds = np.unique(inds)
     return [ l[i] for i in inds ]
 
-def limitInteriorPointsInterpolating(l,numInteriorPoints):
-    '''Like limitInteriorPoints, but interpolates evenly instead; this also means it never clips'''
-    l=np.array(l)
-    if l.ndim==1:
-        l=l[None,:]
-    return [ [ np.interp(ind, range(len(l)), i)
-               for i in l.transpose() ]
-             for ind in np.linspace(0,len(sc),numInteriorPoints+2) ]
+# The active version of this function is defined below (this version is broken/ possibly never finished?)
+#def limitInteriorPointsInterpolatingBAD(l,numInteriorPoints):
+#    '''Like limitInteriorPoints, but interpolates evenly instead; this also means it never clips'''
+#    l=np.array(l)
+#    if l.ndim==1:
+#        l=l[None,:]
+#    return [ [ np.interp(ind, range(len(l)), i)
+#              for i in l.transpose() ]
+#            for ind in np.linspace(0,len(sc),numInteriorPoints+2) ]
 
 def partitionNumpy(l,n):
     '''Like partition, but always clips and returns array, not list'''
@@ -36,6 +51,29 @@ def partitionNumpy(l,n):
     a.resize(len(l)//n,n)
     return a
 
+def addBorder(arr,borderValue=0,borderThickness=1,axis=None):
+    '''For an ND array, create a new array with a border of specified
+       value around the entire original array; optional thickness (default 1)'''
+    # a slice of None or a single int index; handles negative index cases
+    allOrOne = ( slice(None) if axis==None else
+               ( axis        if axis>=0 else 
+                 arr.ndim+axis ))
+    # list of border thickness around each dimension (as an array)
+    tArr = np.zeros(arr.ndim)
+    tArr[allOrOne] = borderThickness
+    # a new array stretched to accommodate the border; fill with border Value
+    arrNew = np.empty( 2*tArr+arr.shape, dtype=arr.dtype )
+    arrNew[:]=borderValue
+    # a slice object that cuts out the new border
+    if axis==None:
+        sl = [slice(borderThickness,-borderThickness)]*arr.ndim
+    else:
+        sl = [slice(None)]*arr.ndim
+        sl[axis] = slice(borderThickness,-borderThickness)
+    # use the slice object to set the interior of the new array to the old array
+    arrNew[sl] = arr
+    return arrNew
+
 def shape_multiply(arr,shapeMultiplier, oddOnly=False, adjustFunction=None):
     '''Works like tile except that it keeps all like elements clumped\n'''
     '''Essentially a non-interpolating multi-dimensional image up-scaler'''
@@ -81,8 +119,22 @@ def shape_multiply_zero_fill(arr,shapeMultiplier):
         #c[tuple([(i//2) for i in shm])]=1
     return shape_multiply(arr,shapeMultiplier,oddOnly=True,adjustFunction=zeroFill)
 
+def interpNaNs(a):
+    '''Changes the underlying array to get rid of NaN values
+       If you want a copy instead, just use interpNaNs(a.copy())
+       Adapted from http://stackoverflow.com/questions/6518811/interpolate-nan-values-in-a-numpy-array'''
+    nans = np.isnan(a)
+    nz_1st = lambda z: z.nonzero()[0]
+    a[nans] = np.interp( nz_1st(nans), nz_1st(~nans), a[~nans] )
+    return a
+
+# A couple working but not-that-great interpolation functions:
 def interpNumpy(l,index):
-    '''Just like interp except that it uses numpy instead of lists.'''
+    '''Just like interp except that it uses numpy instead of lists.
+       If both l and index are integer, this function will also return integer values.
+       The more canonical way to this operation would be:
+           np.interp(index,range(len(l)),l)
+       (where l is already an array)'''
     l = np.array(l)
     m = index % 1
     if m==0:
@@ -124,7 +176,7 @@ def limitInteriorPointsInterpolating(l,numInteriorPoints):
     '''Like limitInteriorPoints, but interpolates evenly instead; this also means it never clips'''
     l=np.array(l)
     return [ interpNumpy(l,ind)
-             for ind in np.linspace(0,len(l)-1,numInteriorPoints+2) ]
+            for ind in np.linspace(0,len(l)-1,numInteriorPoints+2) ]
 
 # Thanks Wikipedia!!
 # A line-drawing algorithm by the pixel...
@@ -176,14 +228,110 @@ def BresenhamFunction(p0,p1): # Generalization to n-dimensions
                 err[i] += delta[imax]
     return l
 
+def linearTransform( points,newRefLine,mirrored=False ):
+    '''Transforms points from a (0,0) -> (1,0) reference coordinate system
+       to a (x1,y1) -> (x2,y2) coordinate system.
+       Optional mirroring.'''
+    newRefLine = np.array(newRefLine)
+    dx,dy = (newRefLine[1]-newRefLine[0])
+    mir = (-1 if mirrored else 1)
+    return np.dot( points, [[dx,dy],[-dy*mir,dx*mir]] ) + newRefLine[0]
+
+def reverseLinearTransform(points,oldRefLine,mirrored=False):
+    '''Transforms points from a (x1,y1) -> (x2,y2) reference coordinate system
+       to a (0,0) -> (1,0) coordinate system.
+       Optional mirroring.'''
+    oldRefLine = np.array(oldRefLine)
+    dx,dy = (oldRefLine[1]-oldRefLine[0])
+    points = (points - oldRefLine[0])
+    mir = (-1 if mirrored else 1)
+    return np.dot( points, [[dx,-dy*mir],[dy,dx*mir]] ) * 1./(dx**2+dy**2)
+
+def FindOptimalScaleAndTranslationBetweenPointsAndReference(points,pointsRef):
+    '''Find the (non-rotational) transformation that best overlaps points and pointsRef
+       aka, minimize the distance between:
+       (xref[i],yref[i],...)
+       and
+       (a*x[i]+x0,a*y[i]+y0,...)
+       using linear least squares
+       
+       return the transformation parameters: a,(x0,y0,...)'''
+    # Force to array of floats:
+    points = np.array(points,dtype=np.float)
+    pointsRef = np.array(pointsRef,dtype=np.float)
+
+    # Compute some means:
+    pm     = points.mean(axis=0)
+    prefm  = pointsRef.mean(axis=0)
+    p2m    = (points**2).mean(axis=0)
+    pTpref = (points*pointsRef).mean(axis=0)
+    
+    a = ((   (pm*prefm).sum() - pTpref.sum()   ) /
+         #   -------------------------------     # fake fraction bar...
+         (      (pm**2).sum() - p2m.sum()      ))
+    p0 = prefm - a*pm
+    return a,p0
+    
+    # More traditional version (2 dimensions only):
+    # xm,ym = pm
+    # xrefm,yrefm = prefm
+    # x2m,y2m = p2m
+    # xTxref,yTyref = pTpref
+    # a = 1.*(xm*xrefm - xCxref + ym*yrefm - yCyref) / (xm**2 - x2m + ym**2 - y2m)
+    # x0,y0 = xrefm - a*xm,yrefm - a*ym
+
 def polyArea(points):
     '''This calculates the area of a general 2D polygon
        Algorithm adapted from Darius Bacon's post:
-       http://stackoverflow.com/questions/451426/how-do-i-calculate-the-surface-area-of-a-2d-polygon'''
-    return 0.5*abs(sum( x0*y1 - x1*y0
-                        for ((x0, y0), (x1, y1)) in zip(points, roll(points)) ))
+       http://stackoverflow.com/questions/451426/how-do-i-calculate-the-surface-area-of-a-2d-polygon
+       Updated to a numpy equivalent for speed'''
+    x0,y0 = np.array(points).T # force to numpy array and transpose
+    x1,y1 = np.roll(points,-1,axis=0).T
+    return 0.5*np.abs(np.sum( x0*y1 - x1*y0 ))
+    
+    # Non-numpy version:
+    #return 0.5*abs(sum( x0*y1 - x1*y0
+    #                   for ((x0, y0), (x1, y1)) in zip(points, roll(points)) ))
+
+def polyCentroid(points):
+    '''Compute the centroid of a generic 2D polygon'''
+    x0,y0 = np.array(points).T # force to numpy array and transpose
+    x1,y1 = np.roll(x0,-1), np.roll(y0,-1)     # roll to get the following values
+    c = x0*y1-x1*y0                            # the cross-term
+    area6 = 3.*np.sum(c)                       # 6*area
+    x,y = np.sum((x0+x1)*c), np.sum((y0+y1)*c) # compute the main centroid calculation
+    return x/area6, y/area6
+    
+    # Non-numpy version:
+    #x = sum( (x0+x1)*(x0*y1-x1*y0)
+    #        for ((x0, y0), (x1, y1)) in zip(points, roll(points)) )
+    #y = sum( (y0+y1)*(y0*x1-y1*x0)
+    #        for ((x0, y0), (x1, y1)) in zip(points, roll(points)) )
+    #return x/area6,y/area6
+    
+    # Single function version:
+    #def _centrX(points):
+    #    '''Compute the x-coordinate of the centroid
+    #       To computer y, reverse the order of x and y in points'''
+    #    return sum( (x0+x1)*(x0*y1-x1*y0)
+    #               for ((x0, y0), (x1, y1)) in zip(points, roll(points)) )
+    #return _centrX(points)/area6,_centrX([p[::-1] for p in points])/area6
+
+def pointDistance(point0,point1):
+    deltas = ( np.array(point1) - point0 ) **2
+    return np.sqrt(np.sum(deltas,axis=-1)).tolist()
+
+def polyPerimeter(points,closeLoop=True):
+    '''This calculates the length of a (default closed) poly-line'''
+    if closeLoop:
+        points = np.concatenate([points,[points[0]]])
+    return sum(pointDistance(points[1:],points[:-1]))
 
 def _getMostCommonVal(l):
+    '''Get the most-occuring value in a list.
+       When multiple values occur the same number of times, returns the minimum one
+       Example:
+           _getMostCommonVal([1,2,4,3,4,5,6,3,5]) -> 3'''
     return Counter( l ).most_common()[0][0]
 
 def GetDirectionsOfSteepestSlope(p0,p1,p2):
@@ -217,7 +365,7 @@ def GetDirectionsOfSteepestSlope(p0,p1,p2):
        Our goal is to find for the other dimension (B) wth the greatest relative slope given that dA=0
        We will compare these by looking at the relative values of dB/dw for some w which is related to the the parametric variables s and t
        
-       Starting with da=0, we see that:
+       Starting with dA=0, we see that:
         0 = dA = (A1-A0)*ds + (A2-A0)*dt
        and
         (A1-A0)*ds = -(A2-A0)*dt
@@ -227,7 +375,7 @@ def GetDirectionsOfSteepestSlope(p0,p1,p2):
         
        Now, in the B direction,
         dB = (B1-B0)*ds + (B2-B0)*dt
-       mltiplying by (A1-A0)*(A2-A0) and collecting terms we find that:
+       multiplying by (A1-A0)*(A2-A0) and collecting terms we find that:
         (A1-A0)*(A2-A0) * dB = (A1-A0)*(A2-A0) * (B1-B0)*ds + (A1-A0)*(A2-A0) * (B2-B0)*dt
         (A1-A0)*(A2-A0) * dB = (A2-A0) * (B1-B0) * dw - (A1-A0) * (B2-B0) * dw
         (A1-A0)*(A2-A0) * dB = ((A2-A0)*(B1-B0) - (A1-A0)*(B2-B0)) * dw
@@ -237,7 +385,7 @@ def GetDirectionsOfSteepestSlope(p0,p1,p2):
        So, for any fixed dimension A, the dimension B will have the greatest relative slope (abs value) if it also maximizes the expression:
         abs( A0*(B2-B1) + A1*(B0-B2) + A2*(B1-B0) )
        This expression can be obtained easily using numpy by:
-        abs( dot( A_i , (roll(B_i,1) - roll(B_i,2)) ) )
+        abs( dot( A_i , (roll(B_i,1) - roll(B_i,-1)) ) )
        This is the expression caclulated below as "slopeMatrix".'''
     
     pointsT = np.transpose([p0,p1,p2])
